'''
Created on 27.11.2020

@author: Weatherman
'''

import sys
from pathlib import Path
from glob import glob
from copy import copy
from datetime import datetime, timedelta
import numpy as np
import warnings

from NumpyRadolanReader import NumpyRadolanReader    # Input
from ASCIIGridWriter    import ASCIIGridWriter       # Output



def test_sum2D_with_nan():
    print("### test_sum2D_with_nan() ###\n")
    
    # Creates a 2D array with uninitialized values
    a = np.ndarray((2, 4))
    # Datentyp ist Default: 'float64'
    a.fill(7) # Fill it in with a constant value 7
    # auch: B.fill(numpy.nan)   # ! nur float-Array
    a[1][0] = np.nan
    a[1][1] = np.nan
    print("a =\n", a)
    
    b = np.ndarray((2, 4))
    b.fill(1.)
    b[1][1] = np.nan
    print("b =\n", b)
    
    """
    Since the inputs are 2D arrays, you can stack them along the third axis with
    np.dstack and then use np.nansum which would ensure NaNs are ignored, unless
    there are NaNs in both input arrays, in which case output would also have NaN.
    -> The behavior that NaN is returned when both inputs are NaN
    changed for NumPy version >1.9.0. In the newer versions, zero is returned!
    """
    c = np.nansum(np.dstack((a, b)), 2)
    
    print("c =\n", c)


def test_sum_same_radolan_file():
    print("### test_sum_same_radolan_file() ###\n")
    
    test_data_dir = "/run/media/loki/ungesichert/Testdaten/RADOLAN_bin"
    
    rw = Path(test_data_dir) / "bin/raa01-rw_10000-1109041250-dwd---bin.gz"
    
    radolan_file = rw
    
    row = 450
    
    nrr = NumpyRadolanReader(radolan_file)    # FileNotFoundError
    nrr.read()
    #nrr.print_data()
    f1 = nrr.data
    f2 = copy(f1)
    
    print(f1[row])
    print()
    
    """
    Since the inputs are 2D arrays, you can stack them along the third axis with
    np.dstack and then use np.nansum which would ensure NaNs are ignored, unless
    there are NaNs in both input arrays, in which case output would also have NaN.
    -> The behavior that NaN is returned when both inputs are NaN
    changed for NumPy version >1.9.0. In the newer versions, zero is returned!
    """
    fsum = np.nansum(np.dstack((f1, f2)), 2)
    
    print(fsum[row])
    



class NumpyRadolanAdder:
    '''
    classdocs
    '''
    
    def __init__(self, dt_beg, dt_end, data_path, prod_id, time_res, asc_filename_path):
        '''
        Constructor
        '''
        
        self.out("dt_beg={}, dt_end={}, prod_id='{}', time_res={}"\
                 .format(dt_beg, dt_end, prod_id, time_res))
        
        # In:
        self._dt_beg = dt_beg
        self._dt_end = dt_end
        self._data_path = data_path
        self._prod_id  = prod_id.lower()    # 'SF' -> 'sf'
        self._time_res = time_res
        self._asc_filename_path = asc_filename_path
        
        # determined:
        self._prec = None
        self._interval_minutes = 0    # of sum
        
        # Out:
        self._sum_field    = None
        self._nodata_field = None
        self._b_first_run  = True
        
    
    def __str__(self):
        return self.__class__.__name__
    
    def out(self, s, ok=True):
        ''' Ausgabemethode '''
        
        if ok:
            print("{}: {}".format(self, s))
        else:
            print("{}: {}".format(self, s), file=sys.stderr)
    
    
    
    def run(self):
        self.out("run()")
        
        # RADOLAN: raa01-rw_10000-1708020250-dwd---bin
        # RADKLIM: raa01-yw2017.002_10000-1006010650-dwd---bin
        general_radolan_file_pattern = "raa01-{}*_10000-{{}}-dwd---bin*".format(self._prod_id)
        
        # datetime.timedelta([days[, seconds[, microseconds[, milliseconds[, minutes[, hours[, weeks]]]]]]])
        td_min = timedelta(minutes=self._time_res)
        
        number_of_added_files = 0
        
        dt = self._dt_beg
        while dt <= self._dt_end:
            fn = general_radolan_file_pattern.format( dt.strftime("%y%m%d%H%M") )
            concrete_radolan_file_pattern = str( Path(self._data_path) / fn )
            
            l_radolan_file = glob(concrete_radolan_file_pattern)    # ...bin, ...bin.gz ?
            
            if l_radolan_file:
                self._add(l_radolan_file[0])    # -> self._prec
                number_of_added_files += 1
            else:
                self.out("expected file '{}' doesn't exist".format(concrete_radolan_file_pattern), False)
            
            dt += td_min
        # while
        
        #print(self._sum_field[0])
        #print(self._nodata_field[0])
        
        self.out("{} files added".format(number_of_added_files))
        
        # Imprint NaN values that were NaN during the entire run:
        nan_positions = np.where(self._nodata_field == True)    # find the remaining 'True's
        self._sum_field[nan_positions] = np.nan
        
        ascii_writer = ASCIIGridWriter(self._sum_field, self._prec, self._asc_filename_path)
        ascii_writer.write()
    
    
    
    def _add(self, radolan_file):
        #self.out("_add('{}')".format(radolan_file))
        
        nrr = NumpyRadolanReader(radolan_file)    # FileNotFoundError
        nrr.read()
        cur_data = nrr.data
        
        self._interval_minutes += nrr.interval
        
        # Init 1:
        if self._b_first_run:
            # Return a new array of given shape and type, filled with fill_value.
            self._nodata_field = np.full(cur_data.shape, True)    # all with NOT NaN
            self._prec = nrr.precision    # save one time
        # if
        
        # every occurence of a pixel (for every looped file) switches a 'True' to a 'False':
        self._nodata_field = np.isnan(cur_data)    # all positions which are NaN
        
        
        # Init 2 add:
        if self._b_first_run:
            self._b_first_run = False
            self._sum_field = cur_data
            return
        
        """
        Since the inputs are 2D arrays, you can stack them along the third axis with
        np.dstack and then use np.nansum which would ensure NaNs are ignored, unless
        there are NaNs in both input arrays, in which case output would also have NaN.
        -> The behavior that NaN is returned when both inputs are NaN
        changed for NumPy version >1.9.0. In the newer versions, zero is returned!
        """
        self._sum_field = np.nansum(np.dstack((self._sum_field, cur_data)), 2)
        
    
    
    def get_statistics(self):
        dim = "rows={}, cols={}".format(*self._sum_field.shape)
        _max = np.nanmax(self._sum_field)    # Maximum of the flattened array, ignoring Nan
        _min = np.nanmin(self._sum_field)    # Minimum of the flattened array, ignoring Nan
        total = self._sum_field.size
        valid = np.count_nonzero(~np.isnan(self._sum_field))    # ~ inverts the boolean matrix returned from np.isnan
        nonvalid = np.count_nonzero(np.isnan(self._sum_field))
        
        return dim, _max, _min, self.mean, total, valid, nonvalid
    
    @property
    def mean(self):
        """ nanmean: The arithmetic mean is the sum of the non-NaN elements along the axis
        divided by the number of non-NaN elements.
        When the array has nothing but nan values, it raises a warning:
        /usr/lib64/python3.4/site-packages/numpy/lib/nanfunctions.py:675: RuntimeWarning: Mean of empty slice
        warnings.warn("Mean of empty slice", RuntimeWarning)
        """
        
        # necessary to catch the warning like an exception:
        # To handle warnings as errors simply use this:
        warnings.filterwarnings('error')
        
        # In order to be able to react to this state, the warning is intercepted:
        try:
            return np.nanmean(self._sum_field)
        
        except Warning:
            self.out("Warning catched: return np.nan", False)
            return np.nan
    
    @property
    def interval_minutes(self):
        return self._interval_minutes
    
    
    

if __name__ == '__main__':
    
    #test_sum2D_with_nan()
    #test_sum_same_radolan_file()
    
    data_path = "/run/media/loki/ungesichert/Testdaten/RADOLAN_bin/viele/sf"
    prod_id   = 'SF'
    time_res  = 1440
    asc_filename_path = "/home/loki/radolan2map/tmp/SF_201501010550-201501070550.asc"
    
    dt_beg = datetime(2015, 1, 1,  5, 50)
    dt_end = datetime(2015, 1, 7,  5, 50)
    
    adder = NumpyRadolanAdder(dt_beg, dt_end, data_path, prod_id, time_res, asc_filename_path)
    adder.run()
    
    